(* Abstract Syntax Tree and functions for printing it *)

type op = Add | Sub | Mult | Div | Equal | Neq | Less | Leq | Greater | Geq |
          And | Or | Mod | Member | Exp | Lshift | Rshift | Band | Bxor | Bor

type uop = Neg | Not | Incr | Decr | Bnot

type typ = Int | Bool | Float | Void | Char | String | Object of string

type access_modifier = Private | Public

type formal = {
  fvtyp: typ;
  fvname: string;
}


type expr =
    Literal of int
  | StrLiteral of string
  | ChLiteral of char
  | FLiteral of string 
  | BoolLiteral of bool
  | Id of string
  | Binop of expr * op * expr
  | Unop of uop * expr
  | Assign of expr * expr
  | Call of string * expr list
  | Noexpr 
  | NewObject of string * expr list
  | AccessMember of expr * expr
  | Null

type vdecl = {
  vaccess: access_modifier;
  vtyp: typ;
  vname: string;
  vexpr: expr;
}
type stmt =
    Block of stmt list
  | Expr of expr
  | Return of expr
  | If of expr * stmt * stmt
  | For of expr * expr * expr * stmt
  | While of expr * stmt
  | Break
  | Continue
  | Variable of vdecl
  | LocalVariable of typ * string * expr

type func_decl = {
  faccess: access_modifier;
  ftyp : typ;
  fname : string;
  fformals : formal list;
  fbody : stmt list;
}



type cbody = {
  fields: vdecl list;
  constructors : func_decl list;
  methods : func_decl list;
}

type class_decl = {
  caccess: access_modifier;
  cname : string;
  cbody : cbody;
}

type arr_decl = {
  arrtyp : typ;
  size: int;
}

type program = Program of class_decl list * func_decl


let rec getType expr env = match expr with
    Id(s) 		-> Int
  | Literal(s) 		-> Int
  | FLiteral(f)		-> Float
  | ChLiteral(c)		-> Char
  | StrLiteral(s)		-> String
  | BoolLiteral(b)		-> Bool
  | Noexpr		-> Void
  | Null			-> Void
  | Binop(e1, op, e2) 	-> (match op with
        Equal|Neq|Less|Leq|Greater|Geq|Or|And -> Bool
      | Add|Sub|Mult|Div -> if ((getType e1 env)=Float) || ((getType e2 env)=Float)
        then Float
        else Int
      | _ -> Int)
  | Unop(op, e)		-> getType e env
  | Assign(s, e)		-> getType e env
  | AccessMember(e1, e2)	-> Int
  | NewObject(s, el) 	-> Object(s)
  | Call(s, el)	-> Int
  | _ -> Void

(* Pretty-printing functions *)
let string_of_typ = function
    Char     -> "char"
  | Void     -> "void"
  |	Bool  -> "boolean"
  |	Float    -> "float"
  | Int 	  -> "int"
  | Object(_) -> "object" 
  | String -> "string" 
let string_of_op = function
    Add -> "+"
  | Sub -> "-"
  | Mult -> "*"
  | Div -> "/"
  | Equal -> "=="
  | Neq -> "!="
  | Less -> "<"
  | Leq -> "<="
  | Greater -> ">"
  | Geq -> ">="
  | And -> "&&"
  | Or -> "||"
  | Mod -> "%"
  | Member -> "."
  | Exp -> "**"
  | Band -> "&"
  | Bor -> "|"
  | Bxor -> "^"
  | Lshift -> "<<"
  | Rshift -> ">>"

let string_of_uop = function
    Neg -> "-"
  | Decr -> "--"
  | Incr -> "++"
  | Not -> "!"
  | Bnot -> "!"

let string_of_char c = String.make 1 c
let rec string_of_expr = function
    Literal(l) -> string_of_int l
  | FLiteral(l) -> l
  | BoolLiteral(true) -> "true"
  | BoolLiteral(false) -> "false"
  | StrLiteral(s) -> s
  | ChLiteral(c) -> string_of_char c
  | Id(s) -> s
  | Binop(e1, o, e2) ->
    string_of_expr e1 ^ " " ^ string_of_op o ^ " " ^ string_of_expr e2
  | Unop(o, e) -> string_of_uop o ^ string_of_expr e
  | Assign(e1, e2) -> string_of_expr e1 ^ " = " ^ string_of_expr e2
  | Call(f, el) ->
    f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
  | Noexpr -> ""
  | Null -> "null"
  | NewObject(s, args) -> "new " ^ s ^ "(" ^ String.concat ", " (List.map string_of_expr args) ^ ")"
  | AccessMember(e1, e2) -> "member access"

let rec string_of_stmt = function
    Block(stmts) ->
    "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | Expr(expr) -> string_of_expr expr ^ ";\n";
  | Return(expr) -> "return " ^ string_of_expr expr ^ ";\n";
  | If(e, s, Block([])) -> "if (" ^ string_of_expr e ^ ")\n" ^ string_of_stmt s
  | If(e, s1, s2) ->  "if (" ^ string_of_expr e ^ ")\n" ^
                      string_of_stmt s1 ^ "else\n" ^ string_of_stmt s2
  | For(e1, e2, e3, s) ->
    "for (" ^ string_of_expr e1  ^ " ; " ^ string_of_expr e2 ^ " ; " ^
    string_of_expr e3  ^ ") " ^ string_of_stmt s
  | While(e, s) -> "while (" ^ string_of_expr e ^ ") " ^ string_of_stmt s
  | Break -> "break;"
  | Continue -> "continue;"


let string_of_vdecl (t, id) = string_of_typ t ^ " " ^ id ^ ";\n"

let string_of_edecl (t, id, e) = string_of_typ t ^ " " ^ id ^ " = " ^ string_of_expr e ^ ";\n"

let string_of_formal f = string_of_typ f.fvtyp ^ " " ^ f.fvname
let string_of_fdecl fdecl =
  string_of_typ fdecl.ftyp ^ " " ^
  fdecl.fname ^ "(" ^ String.concat ", " (List.map string_of_formal fdecl.fformals) ^
  ")\n{\n" ^
  String.concat "" (List.map string_of_stmt fdecl.fbody) ^
  "}\n"

let string_of_cdecl cdecl = "class " ^ cdecl.cname
let string_of_program = function 
  | Program(classes, main) -> String.concat "" (List.map string_of_cdecl classes) ^ "\n" ^ (string_of_fdecl main)
  | _ -> ""
