/* Ocamlyacc parser for GA$$P*/

%{
open Ast
%}

%token SEMI LPAREN RPAREN LBRACE RBRACE COMMA PLUS MINUS TIMES DIVIDE ASSIGN DOT LBRACKET RBRACKET MOD INCR DECR BREAK CONTINUE SWITCH CASE DEFAULT COLON EXP PUBLIC PRIVATE NULL
%token NOT EQ NEQ LT LEQ GT GEQ AND OR BNOT BOR BXOR BAND LSHIFT RSHIFT NEW
%token RETURN IF ELSE FOR WHILE INT BOOL CHAR FLOAT VOID STRING 
%token <int> LITERAL
%token <bool> BLIT
%token <string> SLIT
%token <char> CHLIT
%token <string> ID FLIT
%token CLASS
%token EOF

%start program
%type <Ast.program> program

%nonassoc NOELSE
%nonassoc ELSE
%right ASSIGN
%left OR
%left AND
%left EQ NEQ
%left LT GT LEQ GEQ
%left PLUS MINUS
%left TIMES DIVIDE MOD
%right NOT DOT

%%

program:
  cdecls fdecl EOF { Program($1, $2) }


acc_mod:
 | PUBLIC {Public}
 | PRIVATE {Private}
decls:
   /* nothing */ { ([], [])               }
 | decls vdecl { (($2 :: fst $1), snd $1) }
 | decls fdecl { (fst $1, ($2 :: snd $1)) }
 /* | decls cdecl { (fst $1, ($2 :: snd $1)) } */

fdecl:
   acc_mod typ ID LPAREN formals_opt RPAREN LBRACE stmt_list RBRACE
     { {
       faccess = $1; 
       ftyp = $2;
	 fname = $3;
	 fformals = List.rev $5;
	 fbody = List.rev $8 } }

/* class body */
cbody:
 {{fields = []; constructors =[]; methods =[];}}
| cbody vdecl {{fields = $2::$1.fields; constructors = $1.constructors; methods = $1.methods}}
| cbody constructor {{fields = $1.fields; constructors = $2::$1.constructors; methods = $1.methods}}
| cbody fdecl {{fields = $1.fields; constructors = $1.constructors; methods = $2::$1.methods}}



constructor:
 | ID LPAREN formals_opt RPAREN LBRACE stmt_list RBRACE {{
   faccess = Public;
   fname = $1;
   ftyp = Object($1);
   fformals = $3;
   fbody = List.rev $6;
 }}

cdecls:
| cdecl_list {List.rev $1}                                                                           

cdecl_list:
| cdecl {[$1]}
| cdecl_list cdecl {$2::$1}

cdecl: 
    acc_mod CLASS ID LBRACE cbody RBRACE {{caccess = $1; cname = $3; cbody = $5}}
    /* CLASS ID LBRACE cbody RBRACE {()} */

formals_opt:
    /* nothing */ { [] }
  | formal_list   { $1 }

formal_list:
    typ ID                   { [{fvtyp = $1; fvname = $2;}]     }
  | formal_list COMMA typ ID { {fvtyp = $3; fvname = $4;} :: $1 }

typ:
    INT   { Int   }
  | CHAR  { Char  }
  | STRING{ String}
  | BOOL  { Bool  }
  | FLOAT { Float }
  | VOID  { Void  }


vdecl_list:
    /* nothing */    { [] }
  | vdecl_list vdecl { $2 :: $1 }

vdecl:
   acc_mod typ ID SEMI { {vaccess = $1; vtyp = $2; vname = $3; vexpr = Noexpr;} }
   | acc_mod typ ID ASSIGN expr SEMI {{vaccess = $1; vtyp = $2; vname = $3; vexpr = $5;} }

stmt_list:
    /* nothing */  { [] }
  | stmt_list stmt { $2 :: $1 }

stmt:
    expr SEMI                               { Expr $1               }
  | RETURN expr_opt SEMI                    { Return $2             }
  | LBRACE stmt_list RBRACE                 { Block(List.rev $2)    }
  | IF LPAREN expr RPAREN stmt %prec NOELSE { If($3, $5, Block([])) }
  | IF LPAREN expr RPAREN stmt ELSE stmt    { If($3, $5, $7)        }
  | FOR LPAREN expr_opt SEMI expr SEMI expr_opt RPAREN stmt
                                            { For($3, $5, $7, $9)   }
  | WHILE LPAREN expr RPAREN stmt           { While($3, $5)         }
  | BREAK SEMI                                  {Break}
  | CONTINUE SEMI                               {Continue}
  | vdecl                                   {Variable($1)}
  | typ ID SEMI                             {LocalVariable($1, $2, Noexpr)}
  | typ ID SEMI ASSIGN expr SEMI            {LocalVariable($1, $2, Noexpr)}


expr_opt:
    /* nothing */ { Noexpr }
  | expr          { $1 }

expr:
    LITERAL          { Literal($1)            }
  | CHLIT            { ChLiteral($1)              }
  | SLIT             { StrLiteral($1)               }
  | FLIT	           { FLiteral($1)           }
  | BLIT             { BoolLiteral($1)            }
  | ID               { Id($1)                 }
  | NULL             { Null}
  | expr PLUS   expr { Binop($1, Add,   $3)   }
  | expr MOD    expr { Binop($1, Mod,   $3)   }
  | expr MINUS  expr { Binop($1, Sub,   $3)   }
  | expr TIMES  expr { Binop($1, Mult,  $3)   }
  | expr DIVIDE expr { Binop($1, Div,   $3)   }
  | expr EQ     expr { Binop($1, Equal, $3)   }
  | expr NEQ    expr { Binop($1, Neq,   $3)   }
  | expr LT     expr { Binop($1, Less,  $3)   }
  | expr LEQ    expr { Binop($1, Leq,   $3)   }
  | expr GT     expr { Binop($1, Greater, $3) }
  | expr GEQ    expr { Binop($1, Geq,   $3)   }
  | expr BAND    expr { Binop($1, Band,   $3)   }
  | expr BOR     expr { Binop($1, Bor,    $3)   }
  | expr BXOR     expr { Binop($1, Bxor,    $3)   }
  | expr AND    expr { Binop($1, And,   $3)   }
  | expr OR     expr { Binop($1, Or,    $3)   }
  | expr LSHIFT expr { Binop($1, Lshift,   $3)   }
  | expr RSHIFT expr { Binop($1, Rshift,    $3)   }
  | expr DOT expr { AccessMember($1, $3)}
  | MINUS expr %prec NOT { Unop(Neg, $2)      }
  | expr INCR     { Unop(Incr, $1)      }
  | expr DECR     { Unop(Decr, $1)      }
  | NOT expr         { Unop(Not, $2)          }
  | BNOT expr         { Unop(Bnot, $2)          }
  | expr ASSIGN expr   { Assign($1, $3)         }
  | ID LPAREN args_opt RPAREN { Call($1, $3)  }
  | LPAREN expr RPAREN { $2                   }
  | NEW ID LPAREN args_opt RPAREN {NewObject($2, $4)}

args_opt:
    /* nothing */ { [] }
  | args_list  { List.rev $1 }

args_list:
    expr                    { [$1] }
  | args_list COMMA expr { $3 :: $1 }
