(* Semantically-checked Abstract Syntax Tree and functions for printing it *)

open Ast

type sexpr = 
    SLiteral of int
  | SFLiteral of string
  | SStrLiteral of string
  | SChLiteral of char
  | SBoolLiteral of bool
  | SId of string * typ
  | SBinop of sexpr * op * sexpr * typ
  | SUnop of uop * sexpr * typ
  | SAssign of sexpr * sexpr * typ
  | SCall of string * sexpr list * typ 
  | SNoexpr
  | SNull
  | SAccessMember of sexpr * sexpr * typ
  | SNewObject of string * sexpr list * typ

type sformal = {
  sfvtyp: typ;
  sfvname: string;
}


type svdecl = {
  svaccess: access_modifier;
  svtyp:  typ;
  svname: string;
  svexpr: sexpr;
}

type sstmt =
    SBlock of sstmt list
  | SExpr of sexpr * typ
  | SReturn of sexpr * typ
  | SIf of sexpr * sstmt * sstmt
  | SFor of sexpr * sexpr * sexpr * sstmt
  | SWhile of sexpr * sstmt
  | SBreak
  | SContinue
  | SVariable of svdecl
  | SLocalVariable of typ * string * sexpr

type sfunc_decl = {
  sfaccess: access_modifier;
  sftyp : typ;
  sfname : string;
  sfformals : sformal list;
  sfbody : sstmt list;
}


type scbody = {
  sfields: svdecl list;
  sconstructors: sfunc_decl list;
  smethods: sfunc_decl list;
}

type sclass_decl = {
  scaccess: access_modifier;
  scname : string;
  scbody : scbody;
}

type sarr_decl = {
  sarrtyp : typ;
  ssize: int;
}

type sprogram = {
  classes: sclass_decl list;
  main: sfunc_decl;
  functions: sfunc_decl list;
}

(* Pretty-printing functions *)

(* let rec string_of_sexpr (t, e) =
  "(" ^ string_of_typ t ^ " : " ^ (match e with
        SLiteral(l) -> string_of_int l
      | SBoolLiteral(true) -> "true"
      | SBoolLiteral(false) -> "false"
      | SStrLiteral(s) -> s
      | SChLiteral(c) -> string_of_char c
      | SFLiteral(l) -> l
      | SId(s, t) -> string_of_typ t ^ s
      | SBinop(e1, o, e2, t) ->
        string_of_sexpr (t, e1) ^ " " ^ string_of_op o ^ " " ^ string_of_sexpr (t, e2)
      | SUnop(o, e, t) -> string_of_typ t ^ string_of_uop o ^ string_of_sexpr (t, e)
      | SAssign(e, e2, t) -> string_of_sexpr (t,e) ^ " = " ^ string_of_sexpr (t,e2)
      | SCall(f, el, t) ->
        f ^ "(" ^ String.concat ", " (List.map (fun e -> string_of_sexpr(t,e)) el) ^ ")"
      | SNoexpr -> ""
    ) ^ ")" 

let rec string_of_sstmt = function
    SBlock(stmts) ->
    "{\n" ^ String.concat "" (List.map string_of_sstmt stmts) ^ "}\n"
  | SExpr(expr, t) -> string_of_sexpr (t,expr) ^ ";\n";
  | SReturn(expr, t) -> "return " ^ string_of_sexpr (t,expr) ^ ";\n";
  | SIf(e, s, SBlock([])) ->
    "if (" ^ string_of_sexpr (Int,e) ^ ")\n" ^ string_of_sstmt s
  | SIf(e, s1, s2) ->  "if (" ^ string_of_sexpr (Bool, e) ^ ")\n" ^
                       string_of_sstmt s1 ^ "else\n" ^ string_of_sstmt s2
  | SFor(e1, e2, e3, s) ->
    "for (" ^ string_of_sexpr (Int, e1)  ^ " ; " ^ string_of_sexpr (Bool,e2) ^ " ; " ^
    string_of_sexpr (Int, e3)  ^ ") " ^ string_of_sstmt s
  | SWhile(e, s) -> "while (" ^ string_of_sexpr (Bool,e) ^ ") " ^ string_of_sstmt s
  | SContinue -> "continue;"
  | SBreak -> "break;"

let string_of_sformal = function 
| {sfvtyp = sfvtyp; sfvname = sfvname;} -> string_of_typ sfvtyp ^ " " ^ sfvname
| _ -> ""
let string_of_sfdecl fdecl =
  string_of_typ fdecl.sftyp ^ " " ^
  fdecl.sfname ^ "(" ^ String.concat ", " (List.map string_of_formal fdecl.sfformals) ^
  ")\n{\n" ^
  String.concat "" (List.map string_of_sstmt fdecl.sfbody) ^
  "}\n" *)

let string_of_sprogram sp = match sp with 
| {classes; main; functions;} -> "sprogram"
| _ -> ""
